/*  $Id: ExecRequest.java 238 2006-04-30 20:07:59Z sdevaux $
 *
 * grant - light computationnal grid powered by ants
 * Tabs = 4 chars
 * copyright (C) 2004 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Use of this software with non-GPL software requires specific licencing.
 * contact the author when such usage occures.
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.grant.core;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * Execute request descriptor.
 * handler for all information relating to a task execution request.
 * @version $Id: ExecRequest.java 238 2006-04-30 20:07:59Z sdevaux $
 * @author Sebastien Devaux
 */
public class ExecRequest {
	public static final char PATHDELIM='|';
	/** 
     * create new execute request. 
     * This constructor should be use by Node borkers implementation only as a 
     * way to deserialize exec request.
     * Normale way to create an execute request is to create it from scratch :
     * without any router path.
	 * @param tId task Id.
	 * @param tURL task url.
	 * @param params task "command line" arguments.
	 * @param rp accumulating routing path.
	 * @param s status code 
     */
	public ExecRequest(String tId,String tURL,
							String params,String rp,String s) {
		taskId=tId;						
		taskURL=tURL;
		args=params;
		routerPath=splitRouterPath(rp);
		status=Integer.parseInt(s);
	}
    /**
     * create new execute request. 
	 * @param tURL task url.
	 * @param params task "command line" arguments.
     */
    public ExecRequest(String tURL,String params) {
   		taskId=idSeed+"-"+getIdCount();
		taskURL=tURL;
        args=params;
        routerPath=new String[0];
		status=ROUTE_PROGRESS;
    }
	/**
     * gets task Id.
	 * @return taskId 
     */
	public String getTaskId() {
		return taskId;
	}
	/**
	 * gets task URL.
     * should be called getURI ?
	 * @return taskURL
     */
	public String getTaskURL() {
		return taskURL;
	}
	/** 
	 * gets task arguments.
	 * @return task arguments 
	 */
	public String getArgs() {
		return args;
	}
	/**
	 * @return route status
	 */
	public int getRouteStatus() {
		return status;
	}	
	/**
	 * change route status
	 * @param s new status
	 */
	public void setRouteStatus(int s) {
		status=s;
	}	
	/** 
	 * gets accumulated router path as string.
     * sould be use bye node borker implementation only as a way to
     * serialize router path.
	 * @return route path 
	 */
	public String getRouterPath() {
        StringBuffer sb=new StringBuffer();
        for (int i=0;i<routerPath.length;i++) {
            sb.append(PATHDELIM);
            sb.append(routerPath[i]);
        }
		return sb.toString();
	}
	/** 
	 * Split a string router path as an array.
     * @param rp string to split.
	 * @return splitted router path. 
	 */
	protected String[] splitRouterPath(String rp) {
		Vector v=new Vector();
		StringTokenizer st=new StringTokenizer(rp,""+PATHDELIM);
		while (st.hasMoreTokens()) {
			v.add(st.nextElement());
		}
		String[] res=new String[v.size()];
		return (String[])v.toArray(res);
	}
	/** 
	 * append node to route path. 
	 * @param node node to append. 
     * @return new ExecRequest with node appended.
	 */
	public ExecRequest appendNode(Node node) {
        ExecRequest res=new ExecRequest();
        res.taskId=taskId;
        res.taskURL=taskURL;
        res.args=args;
        res.routerPath=new String[routerPath.length+1];
        int i;
        for (i=0;i<routerPath.length;i++) {
            res.routerPath[i]=routerPath[i];
        }
        res.routerPath[i]=node.getName();
        return res;
	}	
    /**
     * get a node parent in router path.
     * @param node node whose parentis searched.
     * @return predecessor URL when found, null when not any (node not found or
     * node is root).
     */
    public String getParent(Node node) {
        String parent=null;
        for (int i=0;i<routerPath.length;i++) {
            if (node.getName().equals(routerPath[i])) {
                return parent;
            }
			parent=routerPath[i];
        }
        return null;
    }
    /**
     * get a node child in router path.
     * @param node node whose parentis searched.
     * @return successor URL when found, null when not any (node not found or
     * node is leaf).
     */
    public String getChild(Node node) {
        String child=null;
        for (int i=routerPath.length-1;i>=0;i--) {
            if (node.getName().equals(routerPath[i])) {
                return child;
            }
			child=routerPath[i];
        }
        return null;
    }
    /**
     * checks if a node is present in routing path.
     * @param node node to check
     * @return tru when node is found in router path, false when not found
     */
    public boolean member(Node node) {
        for (int i=0;i<routerPath.length;i++) {
            if (node.getName().equals(routerPath[i])) {
                return true;
            } 
        }
        return false;
    }
    /**
     * gets router path length
     * @return node count.
     */
    public int getLength() {
        return routerPath.length;
    }
	/**
	 * gets root node name
	 * @return root node name or null node path is empty.
	 */
	public String getRootNodeName() {
		if (getLength()>0) {
			return routerPath[0];
		}	
		else {
			return null;
		}
	}	
	//--------------------------------------------------------------
	// PRIVATE METHODS
    /**
     * Default empty constructor.
     * privatized to avoid external use.
     */
    private ExecRequest() {
    }
    /**
     * computes and return next id count.
     * @return next id count.
     */
    private static synchronized int getIdCount() {
        return idCount++;        
    }        
	//--------------------------------------------------------------
	// PRIVATE PROPERTIES
	/** 
	 * task ID, this ID is created by first requester. It is
	 * something like requester_hostname.counter, should be unique all over 
	 * the grant.
	 */
	private String taskId;
	/** tasklet to run (URL where to donwload tasklet). */
	private String taskURL;
	/** tasklet command arguments. */
	private String args;
	/** routing path : accumulating while routing. */
	private String[] routerPath;
	/** routing status */
	private int status;
    /** task id prefix.
     * Must be unique for each JVM running nodes. Since prefix includes hostname
     * and time stamp, may not be unique when 2 ore more JVM are started exactly 
     * at the same time on the same host...
     */
    private static String idSeed;
    /** task id count */
    private static int idCount = 0;
    // --------------------------------------------------
    // Static inits
    static {
		String hostname="";
        try {
            hostname=InetAddress.getLocalHost().getHostName();    
        } 
        catch (UnknownHostException ex) {
            hostname="localhost";
        }
        idSeed=hostname+"-"+System.currentTimeMillis();
		idSeed=""+idSeed.hashCode();
    }        
	// routing status consts
	public final static int ROUTE_PROGRESS=	0;
	public final static int ROUTE_OK=		1;
	public final static int ROUTE_OVERLOAD=	2;
	public final static int ROUTE_DELAYED=	3;
}
