/*  $Id: LocalNode.java 238 2006-04-30 20:07:59Z sdevaux $
 *
 * grant - light computationnal grid powered by ants
 * Tabs = 4 chars
 * copyright (C) 2004 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Use of this software with non-GPL software requires specific licencing.
 * contact the author when such usage occures.
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.grant.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import net.eduvax.ai.ant.AntEngine;
import net.eduvax.ai.ant.Marker;
import net.eduvax.ai.ga.Evaluator;
import net.eduvax.ai.ga.RouletteWheele;
import net.eduvax.ai.ga.Selector;

import net.eduvax.util.SLogger;
import net.eduvax.util.Logger;
/**
 * Local node core.
 * Local work impolmentation : routing or launching tasklet when routed to 
 * itself.
 * @version $Id: LocalNode.java 238 2006-04-30 20:07:59Z sdevaux $
 * @author Sebastien Devaux
 */
public class LocalNode extends Node implements Evaluator,Marker {
	/** default delay for ant engine. */
	public static final int ANT_DEF_DELAY = 5000;
	//-----------------------------------------------------------------
	// PUBLIC METHODS
	/**
	 * Handle exec request.
	 * Handling can be :<ul>
	 *	<li>Launche task</li>
	 *	<li>route task to another node</li></ul>
	 * @param er execute request.
	 * @return modified execute request : local node is added to the routing
     *  path.
	 */
	public ExecRequest submitRequest(ExecRequest er) {
		Node node=null;
		synchronized (candidate) {
			candidate.removeAllElements();
			Iterator it=activeNodes.values().iterator();
			while (it.hasNext()) {
				node=(Node)it.next();
				if (!er.member(node)) {
					candidate.add(node);
				}
			}
			node=(Node)selector.select();
		}
		// add this node to the route path
		ExecRequest routed=er.appendNode(this);
		if (node==this) {
			if (checkLoad()) {
	            // schedule tasklet execution 
				taskletManager.schedule(routed); 
				// TODO get shedule return code to report shedule success or error.
				logger.info("Request ["
					+er.getTaskId()+" "
					+er.getTaskURL()+"] scheduled");
				sendEvent(TaskletEvent.SCHEDULED,new TaskletEvent(er,""));
				routed.setRouteStatus(ExecRequest.ROUTE_OK);	
			}
			else {
                if (routed.getParent(this)==null) {
                    submitDelayedRequest(er);    
                }        
			    routed.setRouteStatus(ExecRequest.ROUTE_OVERLOAD);
			}
			return routed;
		}
		else {
			routedTasks++;
			// route request to choosen router
            ExecRequest full=node.submitRequest(routed);
            // feedback routed node.
            antEngine.feedback(node,getFeedback(full));
			// if root node, check route status
			logger.info("Request ["
					+er.getTaskId()+" "
					+er.getTaskURL()+"] routed to "
					+node.getName());
			if (full.getParent(this)==null &&
    				full.getRouteStatus()==ExecRequest.ROUTE_OVERLOAD) {
				submitDelayedRequest(er);
			}
            // return full route.
			return full;
		}
	}
	/**
	 * check for overload.
	 * also adjust minimum delay before new submition attempt.
	 */
	protected boolean checkLoad() {
		Runtime rt=Runtime.getRuntime();
		// System.gc() here is required :
		// Since memory test result will be used to report tasklet schedule
		// memory may never "overflow enough" to trigger garbage collection.
		// So it is manually triggered to collect unused memory and let it 
		// become available for next tasklet.
		System.gc();				
		long free=rt.maxMemory()-rt.totalMemory()+rt.freeMemory();		
		boolean check=(load()<maxLoad)&&(free>(rt.maxMemory()/5));
		if (check) {
			minDelay=(1000+minDelay)/2;
		}
		else {
			minDelay+=1000;
		}	
		return check;
	}
	/**
	 * @return delay before new submition.
	 */
	private int submitDelayedRequest(ExecRequest er) {
        synchronized (this) {
            waitingTasks++;    
        }        
        int delay=random.nextInt();
        delay=delay<0?(-delay)%1000+minDelay:delay%1000+minDelay;    
		er.setRouteStatus(ExecRequest.ROUTE_DELAYED);
    	logger.info("Request ["
						+er.getTaskId()+" "
						+er.getTaskURL()+"] delayed "
						+delay+"ms (final node overload).");
		sendEvent(TaskletEvent.DELAYED,new TaskletEvent(er,""+delay));
        Timer timer=new Timer();
        timer.schedule(new DelayedSubmition(this,er),delay);
		return delay;
	}
	/**
	 * manage end of execution notification.
	 * Updates known nodes description according to excution report
	 * @param er tasklet execution report.
	 * @param len position in routing stack. (&lt;=0 means last) 
	 * @return completion code, 0 is OK.
	 */
	public int execReport(ExecReport er) {
        String child=er.getRequest().getChild(this);
        String parent=er.getRequest().getParent(this);
		if (child==null) {
			runnedTasks++;
		}
        else {
            // feedback reporter
			antEngine.feedback(getNode(child),getFeedback(er));
		}
			
		if (parent==null) {
			execReportToClient(er);
		}
    	else {
	    	// feedback reporter and report to parent in route stack.
		   	getNode(parent).execReport(er);
		}
		return 0;
	}
	/**
	 *
	 */
	public int execReportToClient(ExecReport er) {
		logger.info("Report ["
				+er.getRequest().getTaskId()+" "
				+er.getRequest().getTaskURL()+"] available");
		eventDispatcher.completed(new TaskletEvent(er.getRequest(),er));
		return 0;
	}
	/**
	 * add a node. 
	 * @param nName node's name (xmlrpc server url)
     * @return the new node handler.
	 */
	public Node addNode(String nName) {
        Node n=getNode(nName);
        Node test=(Node)activeNodes.get(nName);
        if (test==null) {
            activeNodes.put(nName,n);
			logger.info("Node "+nName+" added to active node set");
        }
        return n;
    }
    /**
     * Get node if already known, create node handler when not known.
     * @param nName node name.
     * @return node.
     */
    private Node getNode(String nName) {
        Node n=null;
        synchronized (knownNodes) {
            n=(Node)knownNodes.get(nName);
            if (n==null) {
                n=createNode(nName);
                knownNodes.put(nName,n);
                try {
                    PrintStream ps=new PrintStream(
                                new FileOutputStream(knownNodesFile,true));
                    ps.println(nName);
                }    
                catch (IOException ex) {
                    // TODO Log write errors    
					logger.warn("Can't update know nodes file : "+knownNodesFile
							+" ["+ex.getMessage()+"].");
                }        
            }
        }
        return n;
    }
    /**
     * Saves nodes set from rc file.
	 * @param nodeSet node set.
	 * @param nodeFile node file.
     */  
    private void saveNodes(Hashtable nodeSet, File nodesFile) {
		synchronized (nodeSet) {
	        try {
				PrintStream out=new PrintStream(
                            new FileOutputStream(nodesFile));
				Enumeration it=nodeSet.keys();
				while (it.hasMoreElements()) {
					Object key=it.nextElement();
					if (this!=nodeSet.get(key)) {
						// Local node do not save itself !
						out.println(key);
					}
				}
		    }
	        catch (IOException ex) {
				logger.warn("Can't update nodes rc file : "+nodesFile
						+" ["+ex.getMessage()+"].");
	        }        
		}
    }        
    /**
     * Reload known nodes from rc file.
	 * @param nodeSet node set.
	 * @param nodeFile node file.
     */  
    protected void loadNodes(Hashtable nodeSet, File nodesFile) {
		synchronized (nodeSet) {
	        try {
		        String line;    
			    BufferedReader reader=new BufferedReader(
                        new InputStreamReader(    
                            new FileInputStream(nodesFile)));
	            do {
		            line=reader.readLine();
			        if (line!=null&& !line.equals(getName()) ) {
				        Node n=getNode(line);
						if (n!=null) {
					        nodeSet.put(line,n);    
				        }        
					}        
	            } while (line!=null);        
		    }
			catch (IOException ex) {
				logger.warn("Can't read nodes rc file : "+nodesFile
					+" ["+ex.getMessage()+"].");
	        }        
		}	
    }        
	/**
	 * Estimate instant available host power.
	 * higher value means less charge. Should be meaningfull regarding 
	 * host type/cpu freq, all over the node network.
	 * Should be specialized according each runing host since java does not
	 * have any host cpu spec getter, native code may be needed.
	 * Default is running tasks count. 
	 * @return load indicator
	 */
	public float load() {
		return taskletManager.getRunningCount();
	}
	/**
	 * internal state change request.
	 * @param newState next state.
	 * @return current state after request completion. Since requested change
	 * may not be allowed, returned state may be different from newState.
	 */
	private int setState(int newState) {
		boolean ignore=false;
		synchronized(this) {
			switch (state) {
				case STATE_STARTING:
					if (newState==STATE_RUNNING) {
						state=newState;
					}
					else {
						ignore=true;
					}
					break;
				case STATE_SHUTDOWN:
					if (newState==STATE_HALTED) {
						state=newState;
					}
					else {
						ignore=true;
					}
					break;
				default:
					state=newState;
			}	
		}
		if (ignore) {
			logger.warn("Wrong state change request from "
				+STATES_NAME[state]+" to "
				+STATES_NAME[newState]+" ignored.");
		}
		else {
			logger.info("Node state is now "+STATES_NAME[state]+".");
		}
		return state;
	}
	/**
	 * send envent to root node.
	 * @param eventType event type (see TaskletEvent constants)
	 * @param event event to send
	 */
	public void sendEvent(int eventType,TaskletEvent event) {
		String rootNodeName=event.getExecRequest().getRootNodeName();
		if (rootNodeName!=null) {
			Node rootNode=getNode(rootNodeName);
			if (rootNode!=null) {
				rootNode.handleEvent(eventType,event);				
			}	
		}			
	}	
	/**
	 * dispatch events to registered listeners.
	 * @param eventType event type (see TaskletEvent constants)
	 * @param event event to send
	 */ 
	public void handleEvent(int eventType,TaskletEvent event) {
		switch (eventType) {
			case TaskletEvent.SCHEDULED:
				eventDispatcher.scheduled(event);
				;;
			case TaskletEvent.DELAYED:
				eventDispatcher.delayed(event);
				;;
			case TaskletEvent.RUNTIME:
				eventDispatcher.runtimeEvent(event);
				;;
			case TaskletEvent.CANCELLED:
				eventDispatcher.cancelled(event);
				;;
			case TaskletEvent.COMPLETED:
				// this state should not occur here, since this event
				// normally dispatched from "ReportToClient" method.
				eventDispatcher.cancelled(event);
				;;
		}		
	}		
    /**
     * shutdown node
     */
    public void shutdown() {
		logger.info(getName()+" shutdown requested");
		setState(STATE_SHUTDOWN);
		taskletManager.stopAll();	
        antEngine.stop();
		saveNodes(knownNodes,knownNodesFile);	
		saveNodes(activeNodes,activeNodesFile);	
		logger.info(getName()+" shutdown completed");
		setState(STATE_HALTED);
    }
    /**
     * change max load.
     * @param max new maximum load to set.
     */
    public void setMaxLoad(int max) {
        maxLoad=max;   
    }            
    // ---------- Other Node methods implementation
    public String[] getKnownNodes() {
        Object[] values=knownNodes.values().toArray();
        String[] res=new String[values.length];
        for (int i=0;i<values.length;i++) {
            res[i]=(String)values[i];
        }
        return res;    
    }        
	//
	public String status() {
		StringBuffer res=new StringBuffer();
		res.append("<?xml version=\"1.0\"?>\n");
		res.append("<grantNode name=\""+getName()
					+"\" status=\""+STATES_NAME[state]
					+"\" load=\""+load()
					+"\" runnedTasks=\""+runnedTasks
					+"\" routedTasks=\""+routedTasks
					+"\" runningTasks=\""+taskletManager.getRunningCount()
					+"\" waitingTasks=\""+waitingTasks
					+"\">\n");
		res.append("<knownNodes>\n");
		Iterator it=knownNodes.values().iterator();
		while (it.hasNext()) {
			res.append("<node name=\""
					+((Node)it.next()).getName()
					+"\"/>\n");
		}			
		res.append("</knownNodes>\n");
		res.append("<activeNodes>\n");
		it=activeNodes.values().iterator();
		while (it.hasNext()) {
			res.append("<node name=\""
					+((Node)it.next()).getName()
					+"\"/>\n");
		}			
		res.append("</activeNodes>\n");
		res.append("</grantNode>\n");
		return res.toString();
	}
	//
	public void removeNode(String nName) {
        activeNodes.remove(nName);
	}
    // ---------- Evaluator implementation 
    public double fitness(Object o) {
        return ((Node)o).getFitness();
    }
    // ---------- Marker implementation 
    public void dissipate(Object o) {
        Node n=(Node)o;
        double fit=n.getFitness();
        n.setFitness((1+fit)/2);
    }
    public void feedback(Object o,float feedback) {
        Node n=(Node)o;
        double fit=n.getFitness();
        n.setFitness(fit+feedback);
    }
	//-----------------------------------------------------------------
	// PROTECTED METHODS
    /**
     * Create new local node
	 * @param n node's name
     */
	public LocalNode(String n) {
		super(n);
		runnedTasks=0;
		routedTasks=0;
		waitingTasks=0;
		activeNodes=new Hashtable();
        loadNodes(knownNodes,knownNodesFile);
		logger.info("Known nodes set restored.");
        loadNodes(activeNodes,activeNodesFile);
		logger.info("Active nodes set restored.");
        synchronized (knownNodes) {
            knownNodes.put(getName(),this);
        }
        synchronized (activeNodes) {
            activeNodes.put(getName(),this);
        }
		candidate=new Vector();
        setFitness(1.0);
		antEngine=new AntEngine(activeNodes.values(),this);
		antEngine.setDelay(ANT_DEF_DELAY);
		selector=new RouletteWheele(candidate,this);
		antEngine.start();
		taskletManager=new TaskletManager(this);
		setState(STATE_RUNNING);
		logger.info(getName()+" ready");
	}
	/**
	 * compute feedback to local ant colony after task execution.
	 * Each node can use its own feedback computation to handle local
	 * parameters only known here. But each feedback should have the same 
	 * goal : return higher value for higher power used/available.
	 * Default implementation just consider task completion time
	 * @param er execute report
	 * @return computed load indice.
	 */ 
	protected float getFeedback(ExecReport er) {
		return 1.0f/(1.0f+er.getNodeLoad());
	}
	/**
	 * compute feedback to local ant colony after task schedule.
	 * Default implementation just consider route path length.
	 * @param er execute request
	 * @return computed load indice.
	 */
	protected float getFeedback(ExecRequest er) {
		return er.getRouteStatus()==ExecRequest.ROUTE_OK
					? 1.0f/(1.0f+er.getLength())
					: 0.0f;
	}
	/**
	 * defines selector.
	 * should be used in constructor, unless selection mechanism need realy
	 * to be changed during node's life
	 * @param s selector to use to perform node choice.
	 */
	protected void setSelector(Selector s) {
		selector=s;
	}
	/** 
	 * gets current task count on this node.
	 * @return the task count.
	 */
	protected int getTaskCount() {
		return taskletManager.getRunningCount();
	}
    //-----------------------------------------------------------------
    // static inits
    static {
        knownNodes=new Hashtable();
    }
	//-----------------------------------------------------------------
	// PRIVATE PROPERTIES
	/** 
     * node selector. 
     * @djlin.relation composition - -
     */
	private Selector selector;
    /**
     * Known nodes.
     * @djin.relation aggregation - * net.eduvax.grant.core.Node
     * Each node ever encountered is kept inside this set.
     */
    private static Hashtable knownNodes;
	/** 
     * Active nodes set.
     * Should contain only nodes from knownNodes.
     * @djin.relation aggregation - * net.eduvax.grant.core.Node
     */
	private Hashtable activeNodes;
	/** 
     * ant engine manages only local environment. i.e. links to known nodes
	 * from this node. Ant engines uses selector to choose a node each time
     * requested.
     * @djin.relation composition - 1
     */
	private AntEngine antEngine;
	/** 
     * candidate node desciptor list. 
     * Should contain only nodes from activeNodes.
     */
	protected Vector candidate;
	/**
	 * tasks completed by this node.
	 */
	private int runnedTasks;
	/**
	 * tasks routed by this node.
	 */
	private int routedTasks;
	/** internal state */
	private int state=STATE_STARTING;
	/**
	 * tasks wainting to be runned by this node.
	 */
	private int waitingTasks; 
	/**
	 * max load for node.
	 */
	private int maxLoad=64; 
	/** 
	 * tasklet manager
	 * @djin.relation composition - 1
	 */
	private TaskletManager taskletManager;
	/** minimum delay between two submition attempt. Autoadjusted. */
	private int minDelay=1000;
    /** known nodes rc file handler */ 
    static private File knownNodesFile;
    /** active nodes rc file handler */ 
    static private File activeNodesFile;
	/** logger */
	static private SLogger logger;
    /** random number generator */
    static private Random random;
	/** tasklet event dispatcher. */
	static private TaskletEventDispatcher eventDispatcher;
	// node states
	public static final int STATE_STARTING=0;
	public static final int STATE_RUNNING=1;
	public static final int STATE_FULL=2;
	public static final int STATE_SHUTDOWN=3;
	public static final int STATE_HALTED=4;
	public static final String[] STATES_NAME={
			"starting",
			"running",
			"full",
			"shutdown",
			"halted"
	};
    // ---------- static inits
    static {
		logger=Logger.getSLogger("net.eduvax.grant");
        knownNodesFile=new File(Node.getRCDir()+"knownNodes");    
        activeNodesFile=new File(Node.getRCDir()+"activeNodes");    
        random=new Random();
		eventDispatcher=new TaskletEventDispatcher();
    }
    // ---------- inner classes
    private class DelayedSubmition extends TimerTask {
        public DelayedSubmition(LocalNode n,ExecRequest er) {
            node=n;
            request=er;    
        }        
        public void run() {
            synchronized (node) {
                node.waitingTasks--;    
            }        
            node.submitRequest(request);        
        }        
        private ExecRequest request;
        private LocalNode node;
    }    
}
