/*  $Id: Node.java 238 2006-04-30 20:07:59Z sdevaux $
 *
 * grant - light computationnal grid powered by ants
 * Tabs = 4 chars
 * copyright (C) 2004 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Use of this software with non-GPL software requires specific licencing.
 * contact the author when such usage occures.
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.grant.core;

import java.io.File;
import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * Node core.
 * @version $Id: Node.java 238 2006-04-30 20:07:59Z sdevaux $
 * @author Sebastien Devaux
 * @uml:link uses net.eduvax.grant.core.TaskletManager
 * @uml:link relay net.eduvax.grant.core.ExecRequest
 * @uml:link relay net.eduvax.grant.core.ExecReport
 */
public abstract class Node extends Object {
	//-----------------------------------------------------------------
	// PUBLIC METHODS
	/**
	 * Handle exec request.
	 * @param er execute request.
	 * @return path to choosen host, null on error.
	 */
	public abstract ExecRequest submitRequest(ExecRequest er);
	/**
	 * manage end of execution notification.
	 * Updates known nodes description according to excution report
	 * @param er tasklet execution report.
	 * @return completion code, 0 is OK.
	 */
	public abstract int execReport(ExecReport er);
	/**
	 * get node name.
	 * node name should be host name (or URL of xml rpc server controlling this
	 * node...)
	 * @return this node's name.
	 */
	public String getName() {
		return name;
	}
	/**
	 * add a node. 
	 * @param nName node's name (xmlrpc server url)
	 */
	public abstract Node addNode(String nName);
	/**
	 * remove a node.
	 * @param nName node's name (xmlrpc server url)
	 */
	public abstract void removeNode(String nName); 
	/**
	 * Handle tasklet event.
	 * Transmit event to attached event dispatcher.
	 * @param eventType event type (see TaskletEvent constants)
	 * @param event event to dispatch.
	 */ 
	public abstract void handleEvent(int eventType,TaskletEvent event);
	/**
	 * status information.
	 * @return xml status information.
	 */
	public abstract String status(); 
	/**
	 * shutdown node.
	 */
	public abstract void shutdown(); 
    /**
     * request for known nodes name list.
     * @return nodes known from this node.
     */
    public abstract String[] getKnownNodes();
	/**
	 * Estimate instant available host power.
	 * higher value means less charge. Should be meaningfull regarding 
	 * host type/cpu freq, all over the node network.
	 * Should be specialized according each runing host since java does not
	 * have any host cpu spec getter, native code may be needed.
	 * Default is running tasks count. 
	 * @return load indicator
	 */
	public abstract float load();
    /**
     * Create new node.
     * search for a factory for given protocole. 
     * Create new local node if not found (TODO this may not be the best
     * default behavior).
     * @param n node's name (URL)
     * @return node
     */
    public static Node createNode(String n) {
        // get protocol par of URL
        StringTokenizer st=new StringTokenizer(n,":");
        String protocol=st.nextToken();
        NodeFactory factory=(NodeFactory)nodeFactories.get(protocol);
        if (factory!=null) {
            return factory.createNode(n);
        }
        else {
            return new LocalNode(n);
        }
    }
    /**
     * Register a factory for a given protocol.
     * @param protocol protocol name
     * @param nf factory to use with this protocol.
     */
    public static void registerNodeFactory(String protocol,NodeFactory nf) {
        nodeFactories.put(protocol,nf); 
    }
	//-----------------------------------------------------------------
	// PROTECTED METHODS
	/**
	 * Create new node.
	 * @param n node's name
	 */
	protected Node(String n) {
		name=n;
	}
    /**
     * gets node fitness
     * @return current fitness value
     */
    protected double getFitness() {
        return fitness;
    }
    /**
     * change node fitness
     * @param f new fitness
     */
    protected void setFitness(double f) {
        fitness=f;
    }
    /**
     * get preferences file dir path
     */ 
    protected static String getRCDir() {
        return grantRCDir;    
    }        
	//-----------------------------------------------------------------
	// PRIVATE PROPERTIES
	/** node name. */
	private String name;
    /** node fitness */
    private double fitness;
    /** 
     * registered node factories.
     * @djin.relation aggregation - * net.eduvax.grant.core.NodeFactory
     */
    private static Hashtable nodeFactories;
    /** rc files dir path */
    private static String grantRCDir;
    // ---------- static inits
    static {
        nodeFactories=new Hashtable();    
        grantRCDir = System.getProperty("user.home")+
                    System.getProperty("file.separator")+
                    ".grant"+
                    System.getProperty("file.separator");
        File rcDir=new File(grantRCDir);
        rcDir.mkdir();
    }    
}
