/*  $Id: TaskletManager.java 238 2006-04-30 20:07:59Z sdevaux $
 *
 * grant - light computationnal grid powered by ants
 * Tabs = 4 chars
 * copyright (C) 2004 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Use of this software with non-GPL software requires specific licencing.
 * contact the author when such usage occures.
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.grant.core;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.jar.JarFile;

import net.eduvax.util.Logger;
import net.eduvax.util.SLogger;

/**
 * @version $Id: TaskletManager.java 238 2006-04-30 20:07:59Z sdevaux $
 * @author Sebastien Devaux
 */
public class TaskletManager {
    /**
     *
     */     
    protected TaskletManager() {
    }        
	/**
	 * create new tasklet manager attached to a node.
	 * @param n node to attache to this tasklet manager.
	 */
	TaskletManager(LocalNode n) {
		node=n;
		runningTasklets=new Hashtable();
	}
	/**
	 *
	 */
	public synchronized int getRunningCount() {
		return runningCount;
	}
	/**
	 * Schedule a job.
	 * @param er job request description.
	 * @return schedule completion code :<ul>
	 *	<li>SCHED_OK : tasklet thread created and started.</li>
	 *  <li>SCHED_OVERFLOW : max runninf thread count reached.</li>
	 *  <li>SCHED_ERR_CREATE : can't create and start new thread.</li>
	 *  <li>SCHED_ERR_ACCESS : can't get tasklet code.</li>
	 * </ul>
	 */
	public int schedule(ExecRequest er) {
		synchronized (this) {
			if (maxRunningCount>0 && runningCount>=maxRunningCount) {
				return SCHED_OVERFLOW;
			}
		}
		Tasklet tasklet=prepareTasklet(er);
		if (tasklet==null) {
			return SCHED_ERR_ACCESS;
		}
        tasklet.setContext(new TaskletContext(er,node));
		TaskletThread tt=new TaskletThread(this,er,tasklet);
		tt.start();
		synchronized (this) {
			runningTasklets.put(er.getTaskId(),tt);
			runningCount++;
		}
		return SCHED_OK;
	}
	public final static int SCHED_OK=0;
	public final static int SCHED_OVERFLOW=1;
	public final static int SCHED_ERR_CREATE=-1;
	public final static int SCHED_ERR_ACCESS=-2;
	/**
	 * stop all running tasklets.
	 */
	public void stopAll() {
		Enumeration it=runningTasklets.elements();
		while (it.hasMoreElements()) {
			TaskletThread tt=(TaskletThread)it.nextElement();
			tt.interrupt();
		}
	}
	/**
	 * Create TaskletInstance from task URL and command args.
	 * @param taskURL tasklet download URL
	 * @param params command line arguments
	 * @return a tasklet, ready to run, null on error.
	 * @warning for now taskURL is just a class name, tasklet download and
	 *		cache is not yet ready.
	 */
	protected static Tasklet prepareTasklet(ExecRequest er) {
		Tasklet tasklet=null;
		String taskURL=er.getTaskURL();
		String params=er.getArgs();
		Exception exception=null;
		try {
            JarFile jar=jarCache.get(taskURL);
            if (jar==null) {
                throw new IOException("Can't get "+taskURL);
            }                    
            // may be necessay to give current class loader as parent class loader 
            // for this new one.
            URL[] classpath=new URL[1];
            classpath[0]=(new File(jar.getName())).toURL();
            ClassLoader classLoader=new URLClassLoader(classpath);
            String taskletClassName=null;
            int sharpIndex=taskURL.indexOf('#');
            if (sharpIndex>0) {
                taskletClassName=taskURL.substring(sharpIndex+1);
            }    
            else {            
                taskletClassName=
                  jar.getManifest().getMainAttributes().getValue("Main-Class");
            }                  
			Class taskletClass=classLoader.loadClass(taskletClassName);
			tasklet=(Tasklet)taskletClass.newInstance();
		}
		catch (ClassNotFoundException ex) {
			exception=ex;
		}
		catch (InstantiationException ex) {
			exception=ex;
		}
		catch (IllegalAccessException ex) {
			exception=ex;
		}
		catch (IOException ex) {
			exception=ex;
		}
		if (exception!=null) {
			logger.warn("Can't prepare Takslet "
					+taskURL
					+" ["+exception+"].");
		}			
		return tasklet;
	}
	/**
	 * Report task completion to attached node
	 * @param report execution report
	 */ 
	private void report (ExecReport report) {
		synchronized(this) {
			runningCount--;
			if (runningCount<0) {
				logger.warn("Warning taskCount on node "
					+ node.getName() + " is lesser than 0.");
				runningCount=0;
			}
			runningTasklets.remove(report.getRequest().getTaskId());
		}
		node.execReport(report); 
	}	
	/** running task count. */
	private int runningCount = 0;
	/** max running task count. */
	private int maxRunningCount = 500;
	/** running tasklets list. */
	private Hashtable runningTasklets;
	/** attached node. */
	private LocalNode node;
    /** tasklet jar files handler. */
    static private TaskletJarCache jarCache;
	/** logger */
	static private SLogger logger;
    // ---------- static inits
    static {
		logger=Logger.getSLogger(TaskletManager.class.getName());
        jarCache=new TaskletJarCache(".");
    }
	// ----------------------------------------------------------------
	/**
	 * Tasklet Thread wrapper.
	 * Generic thread object to run a single tasklet.
	 * @version $Id: TaskletManager.java 238 2006-04-30 20:07:59Z sdevaux $
	 * @author Sebastien Devaux
	 * @djin.link run net.eduvax.grant.core.Tasklet
	 */
	private class TaskletThread extends Thread {
		/**
		 * Create a new Tasklet Thread from given request to be run
		 * by given node.
		 * @param n node to use to run the tasklet
		 * @param request to-run task description
		 */
		TaskletThread(TaskletManager tm,ExecRequest request,Tasklet t) {
			super(request.getTaskURL()+":"+request.getTaskId());
			er=request;
			task=t;
			manager=tm;
		}
		/**
		 * run tasklet.
		 */
		public void run() {
			int cr=0;
			long time=System.currentTimeMillis();
			try {
				task.run();
				cr=task.getCr();
			}
			catch (Exception ex) {
				cr=-1;
			}	
			float load=manager.node.load(); // load computation is included in running time
			time=System.currentTimeMillis()-time;
			ExecReport report=new ExecReport(er,cr,time,load);
			manager.report(report);
		}
		/** 
		 * Execute request. 
		 * @djin.relation aggregation - 1 
		 */
		private ExecRequest er;
		/**
		 * job to run
		 * @djin.relation composition - 1
		 */
		private Tasklet task; 
		/**
		 *
		 */
		private TaskletManager manager; 
	}	
}
